# Changelog

{{!--
Introduction
• This template tries to follow conventional commits format https://www.conventionalcommits.org/en/v1.0.0/
• The template uses regex to filter commit types into their own headings (this is more than just fixes and features headings)
• It also uses the replaceText function in package.json to remove the commit type text from the message, because the headers are shown instead.
• The text 'Breaking:' or 'Breaking changes:' can be located anywhere in the commit.
• The types feat:, fix:, chore:, docs:, refactor:, test:, style:, perf: must be at the beginning of the commit subject with an : on end.
   • They can optionally have a scope set to outline the module or component that is affected eg feat(bldAssess):
• There is a short hash on the end of every commit that is currently commented out so that change log did not grow too long (due to some system's file size limitations).  You can uncomment if you wish [`{{shorthash}}`]({{href}})

Example Definitions
• feat: A new feature
• fix: A bug fix
• perf: A code change that improves performance
• refactor: A code change that neither fixes a bug nor adds a feature
• style: Changes that do not affect the meaning of the code (white-space, formatting, spelling mistakes, missing semi-colons, etc)
• test: Adding missing tests or correcting existing tests
• docs: Adding/updating documentation
• chore: Something like updating a library version, or moving files to be in a better location and updating all file refs
--}}


{{!-- In package.json need to add this to remove label text from the change log output (because the markdown headers are now used to group them).
   NOTES • Individual brackets have been escaped twice to be Json compliant.
         • For items that define a scope eg feat(bldAssess): We remove the 1st bracket and then re-add it so we can select the right piece of text
{
  "name": "my-awesome-package",
  "auto-changelog": {
    "replaceText": {
      "([bB]reaking:)": "",
      "([bB]reaking change:)": "",
      "(^[bB]uild:)": "",
      "(^[bB]uild\\()": "\\(",
      "(^[dD]eprecate:)": "",
      "(^[dD]eprecate\\()": "\\(",
      "(^[fF]eat:)": "",
      "(^[fF]eat\\()": "\\(",
      "(^[fF]ix:)": "",
      "(^[fF]ix\\()": "\\(",
      "(^[cC]hore:)": "",
      "(^[cC]hore\\()": "\\(",
      "(^[dD]ocs:)": "",
      "(^[dD]ocs\\()": "\\(",
      "(^[rR]efactor:)": "",
      "(^[rR]efactor\\()": "\\(",
      "(^[tT]est:)": "",
      "(^[tT]est\\()": "\\(",
      "(^[sS]tyle:)": "",
      "(^[sS]tyle\\()": "\\(",
      "(^[pP]erf:)": "",
      "(^[pP]erf\\()": "\\("
    }
  }
}
 --}}

 {{!--
 Regex reminders
 ^ = starts with
 \( = ( character (otherwise it is interpreted as a regex lookup group)
 * = zero or more of the previous character
 \s = whitespace
 . = any character except newline
 | = or
 [aA] = character a or character A
 --}}


{{#each releases}}
  {{#if href}}
    ## [{{title}}]({{href}})
  {{else}}
    ## {{title}}
  {{/if}}

  > {{niceDate}}

  {{#if summary}}
    {{summary}}
  {{/if}}

  {{! List commits that fix a given issues}}
  {{#each fixes}}
    - {{#if commit.breaking}}**Breaking change:** {{/if}}{{commit.subject}}{{#each fixes}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}}){{/each}}
  {{/each}}

  {{! List merge commits with 'breaking:' or 'Breaking change:' anywhere in the message under a heading}}
  {{#commit-list merges heading='### Breaking Changes :warning:' message='[bB]reaking [cC]hange:|[bB]reaking:' exclude='\[skip-changelog\]'}}
    - {{message}} @{{author}}  <!--[`#{{id}}`]({{href}}) -->
  {{/commit-list}}

  {{! List commits organised under a heading, but not those already listed in the breaking section }}
      {{#commit-list merges heading='### Build' message='^[bB]uild:|^[bB]uild\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Deprecated Features' message='^[dD]eprecate:|^[dD]eprecate\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### New Features' message='^[fF]eat:|^[fF]eat\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Fixes' message='^[fF]ix:|^[fF]ix\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Chores And Housekeeping' message='^[cC]hore:|^[cC]hore\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Documentation Changes' message='^[dD]ocs:|^[dD]ocs\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Refactoring and Updates' message='^[rR]efactor:|^[rR]efactor\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Changes to Test Assests' message='^[tT]est:|^[tT]est\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Tidying of Code eg Whitespace' message='^[sS]tyle:|^[sS]tyle\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### Performance Improvements' message='^[pP]erf:|^[pP]erf\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### CI Improvements' message='^[cC][iI]:|^[cC][iI]\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list merges heading='### General Changes' exclude='[bB]reaking [cC]hange:|[bB]reaking:|^[bB]uild:|^[bB]uild\(|^[dD]eprecate:|^[dD]eprecate\(|^[fF]eat:|^[fF]eat\(|^[fF]ix:|^[fF]ix\(|^[cC]hore:|^[cC]hore\(|^[cC][iI]:|^[cC][iI]\(|^[dD]ocs:|^[dD]ocs\(|^[rR]efactor:|^[rR]efactor\(|^[tT]est:|^[tT]est\(|^[sS]tyle:|^[sS]tyle\(|^[pP]erf:|^[pP]erf\(|\[skip-changelog\]'}}
        - {{message}} ({{author}}{{#if href}} - [#{{id}}]({{href}}){{/if}})
      {{/commit-list}}

  {{! List commits with 'breaking:' or 'Breaking change:' anywhere in the message under a heading}}
  {{#commit-list commits heading='### Breaking Changes :warning:' message='[bB]reaking [cC]hange:|[bB]reaking:' exclude='\[skip-changelog\]'}}
    - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
  {{/commit-list}}

  {{! List commits organised under a heading, but not those already listed in the breaking section }}
      {{#commit-list commits heading='### Build' message='^[bB]uild:|^[bB]uild\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Deprecated Features' message='^[dD]eprecate:|^[dD]eprecate\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### New Features' message='^[fF]eat:|^[fF]eat\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Fixes' message='^[fF]ix:|^[fF]ix\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Chores And Housekeeping' message='^[cC]hore:|^[cC]hore\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Documentation Changes' message='^[dD]ocs:|^[dD]ocs\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Refactoring and Updates' message='^[rR]efactor:|^[rR]efactor\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Changes to Test Assests' message='^[tT]est:|^[tT]est\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Tidying of Code eg Whitespace' message='^[sS]tyle:|^[sS]tyle\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### Performance Improvements' message='^[pP]erf:|^[pP]erf\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### CI Improvements' message='^[cC][iI]:|^[cC][iI]\(' exclude='[bB]reaking [cC]hange:|[bB]reaking:|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

      {{#commit-list commits heading='### General Changes' exclude='[bB]reaking [cC]hange:|[bB]reaking:|^[bB]uild:|^[bB]uild\(|^[dD]eprecate:|^[dD]eprecate\(|^[fF]eat:|^[fF]eat\(|^[fF]ix:|^[fF]ix\(|^[cC]hore:|^[cC]hore\(|^[cC][iI]:|^[cC][iI]\(|^[dD]ocs:|^[dD]ocs\(|^[rR]efactor:|^[rR]efactor\(|^[tT]est:|^[tT]est\(|^[sS]tyle:|^[sS]tyle\(|^[pP]erf:|^[pP]erf\(|\[skip-changelog\]'}}
        - {{subject}} ({{author}}{{#if href}} - [{{shorthash}}]({{href}}){{/if}})
      {{/commit-list}}

{{/each}}
